import numpy as np
"""
Is
"""
class XYClassDistribution:

    def __init__(self, layout, marker_to_class_id=None):
        self.layout = layout
        if marker_to_class_id is None:
            self.marker_to_class_id = {elem: elem for line in layout for elem in line}
        else:
            self.marker_to_class_id = marker_to_class_id
        self.class_layout = self._create_layout(self.layout)

    def _create_layout(self, layout):
        class_layout = []
        for line in layout:
            class_layout.append([])
            for sym in line:
                class_layout[-1].append(self.marker_to_class_id[sym])
        return class_layout

    def __call__(self):
        return np.asarray(self.class_layout)


import numpy as np

def one_hot(i, n):
    v = np.zeros(n)
    v[i] = 1.
    return v

def one_hot_nd(nd_int_array, N=None):

    if N is None:
        N = len(np.unique(nd_int_array))

    oh_mat = []
    for x in np.nditer(nd_int_array):
        oh_mat.append(one_hot(x, N))
    return np.asarray(oh_mat).reshape(nd_int_array.shape + (N,))



import numpy as np
"""
How.
"""

class AbstractDynamics:

    def __init__(self, state_space):
        self.state_space = state_space
        pass

    def actions(self):
        raise NotImplementedError

    def take_action(self, state, action):
        raise NotImplementedError

    def get_next_states_distribution(self, state, action):
        raise NotImplementedError

    def infer_action(self, s1, s2):
        raise NotImplementedError

    def __call__(self, state, action):
        return self.get_next_states_distribution(state, action)

class XYDynamics(AbstractDynamics):
    ACTIONS = ["U", "D", "L", "R"]
    OOPS_ACTIONS = {"U": ["L", "R"], "D": ["R", "L"], "L": ["D", "U"], "R": ["U", "D"]}

    def __init__(self, state_space, slip_prob=0.):
        super().__init__(state_space)
        self.slip_prob = slip_prob
        self.H, self.W = self.state_space.limits
        self.a_to_idx = {a: i for i, a in enumerate(self.ACTIONS)}

    def actions(self):
        return self.ACTIONS

    def action_idx(self, action):
        return self.a_to_idx[action]

    def _next_state(self, state, action):
        loc = state.location

        if action not in self.ACTIONS:
            raise Exception("Invalid action {}!".format(action))

        if action == "U" and loc[0] - 1 >= 0:
            loc_prime = (loc[0] - 1, loc[1])
        elif action == "D" and  loc[0] + 1 < self.H:
            loc_prime = (loc[0] + 1, loc[1])
        elif action == "L" and loc[1] - 1 >= 0:
            loc_prime = (loc[0], loc[1] - 1)
        elif action == "R" and loc[1] + 1 < self.W:
            loc_prime = (loc[0], loc[1] + 1)
        else:
            loc_prime = loc
        # print(state, action, "->", loc_prime)
        return self.state_space.loc_to_state_dict[loc_prime]

    def get_next_states_distribution(self, state, action):
        action_list = [action, ] + self.OOPS_ACTIONS[action]
        p_vals = [ 1. - self.slip_prob, ] + [ self.slip_prob ] * len(self.OOPS_ACTIONS[action])
        return [(self._next_state(state, action), p_vals[idx]) for idx, action in enumerate(action_list)]

    def take_action(self, state, action):
        if state.is_terminal():
            return state
        if self.slip_prob > np.random.random():
            action = np.random.choice(self.OOPS_ACTIONS[action])
        return self._next_state(state, action)

    def _is_valid_loc(self, loc):
        if 0 <= loc[0] < self.H and 0 <= loc[1] < self.W:
            return True
        return False

    def infer_action_by_loc(self, loc1, loc2):
        # TODO: handle self.slip_prob
        if not self._is_valid_loc(loc1) or not self._is_valid_loc(loc2):
            raise Exception("Invalid input: locations out of bound!")
        if loc1[0] == loc2[0] and loc1[1] == loc2[1]:
            raise Exception("Invalid input: action not supported!")
        if loc1[0] != loc2[0] and loc1[1] != loc2[1]:
            raise Exception("Invalid input: action not supported!")

        if loc2[0] - loc1[0] > 0:
            return "D"
        elif loc2[0] - loc1[0] < 0:
            return "U"
        elif loc2[1] - loc1[1] > 0:
            return "R"
        else: # loc2[1] - loc1[1] < 0
            return "L"

    def infer_action(self, s1, s2):
        return self.infer_action_by_loc(s1.location, s2.location)

    def loc_lst_to_a_lst(self, loc_lst):
        return [self.infer_action_by_loc(loc_lst[i], loc_lst[i+1]) for i in range(len(loc_lst)-1)] + [None,]

    def a_lst_to_loc_lst(self, loc_0, a_lst):
        loc_lst = [loc_0]
        loc = loc_0
        for a in a_lst:
            loc = self.take_action(self.state_space.loc_to_state_dict[loc], a).location
            loc_lst.append(loc)
        return loc_lst


import itertools
import numpy as np
# from collections import defaultdict
"""
What, Where.
"""

class DiscreteStateSpace:
    ''' Discrete State Space specification class
    Note: Designed to handle any dimensionality but dim > 2 is not tested.
    '''

    def __init__(self, *args):
        self.n_dims = len(args)
        self.limits = args
        self.n_states = np.product(self.limits)
        self.idxs = self._get_idxs()
        self.space = self._get_space()
        self.state_lst, self.loc_to_state_dict, self.state_to_loc_dict, self.loc_to_idx_dict = self._get_states()

    def _get_idxs(self):
        return np.arange(self.n_states).reshape(self.limits)

    def _get_space(self):
        return np.array(list(itertools.product(*[np.arange(lim) for lim in self.limits]))).reshape(
            self.limits + (self.n_dims,))

    def _get_states(self):
        state_list = []
        loc_to_state_dict = {}
        state_to_loc_dict = {}
        loc_to_idx_dict = {}
        idx = 0
        for loc in list(itertools.product(*[np.arange(lim) for lim in self.limits])):
            state = State(location=loc, idx=idx, parent_state_space=self)
            state_list.append(state)
            loc_to_state_dict[loc] = state
            state_to_loc_dict[state] = loc
            loc_to_idx_dict[loc] = idx
            idx += 1
        return state_list, loc_to_state_dict, state_to_loc_dict, loc_to_idx_dict

    def __str__(self):
        #return "".join(str(s) if ((idx+1) % self.limits[-1] != 0) else str(s) + "\n" for idx, s in enumerate(self.states_lst))
        return str(self.space)

    def print_states_meta(self):
        for s in self.state_lst:
            print(s._meta())

    def at_loc(self, loc):
        return self.loc_to_state_dict[loc]

    def __call__(self, idx=None):
        if idx is None:
            return self.state_lst
        else:
            return self.__getitem__(idx)

    def __getitem__(self, idx):
        if type(idx) == tuple or type(idx) == list:
            return self.at_loc(loc=tuple(idx))
        else:
            return self.state_lst[idx]

    def __len__(self):
        return self.n_states

    def all(self):
        return self.state_lst

    def attach_classes(self, class_ids=[]):
        if len(class_ids) != self.n_states:
            raise Exception("Require class id for each state!")
        self.n_classes = len(np.unique(class_ids))
        for idx, class_id in enumerate(class_ids):
            self.state_lst[idx].attach_class(class_id)

    def sample_and_attach_classes(self, class_ids=[], p_dist=None):
        S = class_ids
        if p_dist is None:
            p_dist = np.ones(len(S)) / len(S)
        if len(p_dist) != len(class_ids):
            raise Exception("class_ids and p_dist must have same length!")
        self.uniq_classes = class_ids
        class_ids = np.random.choice(S, self.n_states, p=p_dist) #.reshape(self.limits)
        for idx, class_id in enumerate(class_ids):
            self.state_lst[idx].attach_class(class_id)

    def override_classes_by_loc(self, loc_lst, class_id_lst):
        for i, loc in enumerate(loc_lst):
            self.loc_to_state_dict[loc].attach_class(class_id_lst[i])
        self.uniq_classes = np.unique(self.class_ids)

    @property
    def class_ids(self):
        return np.asarray([self.state_lst[idx].get_class() for idx in range(self.n_states)])

    @property
    def num_classes(self):
        return len(self.uniq_classes)

    def attach_feature_spec(self, PHI_spec, compute=True):
        for state in self.state_lst:
            state.attach_feature_spec(PHI_spec, compute=compute)

    def add_feature_spec_ref(self, key, ref_key):
        for state in self.state_lst:
            state.add_feature_spec_ref(key, ref_key)

    def clear_feature_spec(self):
        for state in self.state_lst:
            state.clear_feature_spec()

    def features(self, loc=None, idx=None, gridded=False, numpyize=True, key=None):
        if loc is not None:
            return np.asarray(self.loc_to_state_dict[loc].get_features(key=key))
        elif idx is not None:
            return np.asarray(self.state_lst[idx].get_features(key=key))
        else:
            features_lst = [self.state_lst[idx].get_features(key=key) for idx in range(self.n_states)]
            if gridded:
                return np.asarray(features_lst).reshape(self.limits + self.feature_dim(key=key))
            else:
                if numpyize:
                    return np.asarray(features_lst, dtype=np.float32)
                else:
                    return features_lst

    def feature_dim(self, key=None):
        return self.features(idx=0, loc=None, gridded=False, key=key).shape

    def attach_reward_spec(self, R_spec, compute=True):
        for state in self.state_lst:
            state.attach_reward_spec(R_spec, compute=compute)

    def add_reward_spec_ref(self, key, ref_key):
        for state in self.state_lst:
            state.add_reward_spec_ref(key, ref_key)

    def get_reward_keys(self):
        return self.state_lst[0].get_reward_keys()

    def get_reward_spec(self, key):
        return self.state_lst[0].get_reward_spec(key)

    def get_feature_keys(self):
        return self.state_lst[0].get_feature_keys()

    def clear_reward_spec(self):
        for state in self.state_lst:
            state.clear_reward_spec()

    def rewards(self, numpyize=True, gridded=False, key=None):
        if gridded:
            rewards = np.asarray([self.state_lst[idx].get_reward(key=key) for idx in range(self.n_states)],
                                 dtype=np.float32)
            return rewards.reshape(*self.limits, 1)
        else:
            rewards = [self.state_lst[idx].get_reward(key=key) for idx in range(self.n_states)]
            if numpyize:
                return np.asarray(rewards, dtype=np.float32)
            else:
                return rewards

    def set_terminal_status_by_loc(self, loc_lst, b_terminal_status=True):
        for loc in loc_lst:
            self.loc_to_state_dict[loc].set_terminal_status(b_terminal_status)

    def set_terminal_status_by_idx(self, idx_lst, b_terminal_status=True):
        for idx in idx_lst:
            self.state_lst[idx].set_terminal_status(b_terminal_status)

    def get_terminal_states(self):
        return [s for s in self.state_lst if s.terminal_status is True]

    def reset_terminal_status(self):
        for s in self.state_lst:
            s.set_terminal_status(False)

    def _organize_to_grid(self, values):
        return np.asarray(values).reshape(self.shape() + values[0].shape)

    def shape(self):
        return self.limits


class State(object):
    ''' State specification class
    Adapted from https://github.com/david-abel/simple_rl/simple_rl/mdp/StateClass.py
    '''

    def __init__(self, location, idx=None,
                 class_id=None, features=None,
                 reward=None, reward_spec=None,
                 key=None, feature_spec=None,
                 terminal_status=False, parent_state_space=None):
        self.location = location
        self.idx = idx
        self.class_id = class_id
        self.terminal_status = terminal_status
        self.parent_state_space = parent_state_space
        self.features = features
        self.feature_spec_dict = {}
        if feature_spec is not None:
            self.feature_spec_dict[key] = feature_spec
        self.reward = reward
        self.reward_spec_dict = {}
        if reward_spec is not None:
            self.reward_spec_dict[key] = reward_spec

    def is_terminal(self):
        return self.terminal_status

    def set_terminal_status(self, b_terminal_status):
        self.terminal_status = b_terminal_status

    def get_idx(self):
        return self.idx

    def get_id(self):
        return self.location

    def get_location(self):
        return self.location

    def get_class(self):
        return self.class_id

    def attach_class(self, class_id):
        self.class_id = class_id

    def attach_feature_spec(self, feature_spec, compute=True):
        self.feature_spec_dict[feature_spec.get_key()] = feature_spec
        if compute:
            self.compute_features(feature_spec.get_key())

    def add_feature_spec_ref(self, key, ref_key):
        self.feature_spec_dict[ref_key] = self.feature_spec_dict[key]

    def clear_feature_spec(self):
        self.feature_spec_dict = {}

    def attach_reward_spec(self, reward_spec, compute=True):
        self.reward_spec_dict[reward_spec.get_key()] = reward_spec
        if compute:
            self.compute_reward(reward_spec.get_key())

    def add_reward_spec_ref(self, key, ref_key):
        self.reward_spec_dict[ref_key] = self.reward_spec_dict[key]

    def clear_reward_spec(self):
        self.reward_spec_dict = {}

    def get_features(self, key=None, recompute=True):
        if recompute:
            self.compute_features(key=key)
        return self.features

    def get_reward(self, key=None, recompute=True):
        if recompute:
            self.compute_reward(key=key)
        return self.reward

    def compute_reward(self, key=None):
        reward_spec = self.get_reward_spec(key)
        if reward_spec is not None:
            self.reward = reward_spec.compute_state_reward(self)
        else:
            pass

    def compute_features(self, key=None):
        feature_spec = self.get_feature_spec(key)
        if feature_spec is not None:
            self.features = feature_spec.compute_features(self)
        else:
            pass

    def get_reward_keys(self):
        return list(self.reward_spec_dict.keys())

    def get_reward_spec(self, key):
        return self.reward_spec_dict[key]

    def get_feature_keys(self):
        return list(self.feature_spec_dict.keys())

    def get_feature_spec(self, key):
        return self.feature_spec_dict[key]

    def parent_space(self):
        return self.parent_state_space

    def __hash__(self):
        if type(self.location).__module__ == np.__name__:
            # Numpy arrays
            return hash(str(self.location))
        elif self.location.__hash__ is None:
            return hash(tuple(self.location))
        else:
            return hash(self.location)

    def __str__(self):
        return "State: " + str(self.location)

    def _meta(self):
        return "State: " + str(self.location) + " [ "+ \
               ("C {} ".format(self.get_class()) if self.get_class() is not None else "") + \
               ("R {:.2f} ".format(self.get_reward()) if self.get_reward() is not None else "") + \
               ("features {} ".format(self.get_features().shape) if self.get_features() is not None else "") + \
               ("Terminal " if self.is_terminal() else "") + \
                "]"

    def __repr__(self):
        return str(self.__module__) + "." + self.__class__.__name__ + str(self.location) + " at " + hex(id(self))

    def __eq__(self, other):
        if isinstance(other, State):
            return self.location == other.location
        return False

    def __getitem__(self, index):
        return self.location[index]

    def __len__(self):
        return len(self.location)


from abc import ABC, abstractmethod
from .state import State, DiscreteStateSpace
import numpy as np
"""
Objective, Utility
"""

class AbstractStateRewardSpec(ABC):

    def __init__(self, key=None, feature_key=None, preprocess_fn=lambda x: x, postprocess_fn=lambda x: x):
        self.key = key
        self.feature_key = key if feature_key is None else feature_key
        self.preprocess_fn = preprocess_fn
        self.postprocess_fn = postprocess_fn

    def __call__(self, x):
        return self.reward(x)

    def get_key(self):
        return self.key

    def get_feature_key(self):
        return self.feature_key

    @abstractmethod
    def compute_state_reward(self, state):
        return

    @abstractmethod
    def compute_state_space_reward(self, discrete_state_space):
        return

    def reward(self, x):
        if isinstance(x, State):
            return self.compute_state_reward(x)
        elif isinstance(x, DiscreteStateSpace):
            return self.compute_state_space_reward(x)
        else:
            raise ValueError("Incorrect input in reward spec: {}".format(type(x)))

    def preprocess(self, features):
        return self.preprocess_fn(features)

    def postprocess(self, reward):
        return self.postprocess_fn(reward)

    def features(self, state):
        return self.preprocess(state.get_features(key=self.get_feature_key()))

    @abstractmethod
    def get_model(self):
        return


class RewardStateScalar(AbstractStateRewardSpec):

    def __init__(self, loc_to_reward_dict, class_id_to_reward_dict, default=0,
                 key=None, feature_key=None, preprocess_fn=lambda x: x, postprocess_fn=lambda x: x):
        super().__init__(key, feature_key, preprocess_fn, postprocess_fn)
        self.loc_to_reward_dict = loc_to_reward_dict
        self.class_id_to_reward_dict = class_id_to_reward_dict
        self.default = default

    def compute_state_reward(self, state):
        # loc_to_reward_dict overrides class_id_to_reward_dict
        if (self.loc_to_reward_dict is not None) and state.location in self.loc_to_reward_dict:
            return self.postprocess_fn(
                self.loc_to_reward_dict[self.preprocess_fn(state.location)]
            )
        elif (self.class_id_to_reward_dict is not None) and state.class_id in self.class_id_to_reward_dict:
            return self.postprocess_fn(
                self.class_id_to_reward_dict[self.preprocess_fn(state.class_id)]
            )
        else:
            return self.postprocess_fn(
                self.preprocess_fn(self.default)
            )

    def get_model(self):
        raise NotImplementedError

    def compute_state_space_reward(self, discrete_state_space):
        rewards = []
        for state in discrete_state_space:
            rewards.append(self.compute_state_reward(state))
        return rewards

class RewardStateFeatureModel(AbstractStateRewardSpec):

    def __init__(self, r_model, key=None, feature_key=None,
                 preprocess_fn=lambda x: x, postprocess_fn=lambda x: x):
        super().__init__(key, feature_key, preprocess_fn, postprocess_fn)
        self.r_model = r_model

    def compute_state_reward(self, state):
        return self.postprocess_fn(
            self.r_model(self.preprocess_fn(state.get_features(key=self.get_feature_key())))
        )

    def get_model(self):
        return self.r_model

    def compute_state_space_reward(self, discrete_state_space):
        PHI = np.asarray(discrete_state_space.features(loc=None, idx=None,
                                                       gridded=False, numpyize=False,
                                                       key=self.get_feature_key()))
        return self.postprocess_fn(
            self.r_model(self.preprocess_fn(PHI))
        )



import setuptools
exec(open('navigation_mdp/_version.py').read())

# Ref: https://packaging.python.org/tutorials/packaging-projects/

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="navigation_mdp",
    version=__version__,
    author="Yagnesh Revar",
    author_email="mailto.yagnesh+github@gmail.com",
    description="A lightweight library for defining navigation grid world",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/yrevar/navigation_mdp",
    packages=setuptools.find_packages(),
    keywords = ['Markov Decision Process', 'MDP', 'Navigation'],
    install_requires = [
        'navigation_vis'
    ],
    dependency_links=[
        'git+ssh://git@github.com/yrevar/navigation_vis@v0.8#egg=navigation_vis-0.9'
    ],
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: Apache Software License",
        "Operating System :: OS Independent",
    ],
    python_requires='>=3.6',
)



import numpy as np
from navigation_mdp.utils import one_hot_nd
"""
Why.
"""

class AbstractStateFeatureSpec:

    def __init__(self, key=None):
        self.key = key
        # self.features_lst = None

    def __call__(self, state):
        return self.compute_features(state)

    def get_key(self):
        return self.key

    def compute_features(self, state):
        raise NotImplementedError

    # def __call__(self, loc=None, idx=None, gridded=False):
    #     if loc is not None:
    #         return self.get_features_at_loc(loc)
    #     elif idx is not None:
    #         return self.get_features_at_idx(idx)
    #     else:
    #         if gridded:
    #             return self.get_all_features().reshape(self.state_space.limits + self.features_lst.shape[1:])
    #         else:
    #             return self.get_all_features()

    # def __getitem__(self, idx):
    #     return self.get_features_at_idx(idx)
    #
    # def __len__(self):
    #     return len(self.features_lst)
    #
    # def get_all_features(self):
    #     return self.features_lst
    #
    # def get_features_at_idx(self, idx):
    #     return self.features_lst[idx]
    #
    # def get_features_at_loc(self, loc):
    #     return self.features_lst[self.state_space.loc_to_idx_dict[loc]]


class FeatureStateIndicator(AbstractStateFeatureSpec):

    def __init__(self, key=None):
        super().__init__(key)

    def compute_features(self, state):
        return state.get_idx()


class FeatureStateIndicatorOneHot(AbstractStateFeatureSpec):

    def __init__(self, key=None, dim=None):
        super().__init__(key)
        self.dim = dim

    def compute_features(self, state):
        dim = state.parent_space().n_states if self.dim is None else self.dim
        return one_hot_nd(np.asarray(state.get_idx()), N=dim)


class FeatureClassIndicator(AbstractStateFeatureSpec):

    def __init__(self, key=None):
        super().__init__(key)

    def compute_features(self, state):
        return state.get_class()


class FeatureClassIndicatorOneHot(AbstractStateFeatureSpec):

    def __init__(self, key=None, dim=None):
        super().__init__(key)
        self.dim = dim

    def compute_features(self, state):
        dim = max(state.parent_space().class_ids) + 1 if self.dim is None else self.dim
        return one_hot_nd(np.asarray(state.get_class()), N=dim)


class FeatureClassImage(AbstractStateFeatureSpec):

    def __init__(self, feature_map, key=None):
        super().__init__(key)
        self.feature_map = feature_map

    def compute_features(self, state):
        return self.feature_map[state.get_class()]


class FeatureClassImageSampler(AbstractStateFeatureSpec):

    def __init__(self, feature_sampler, key=None):
        super().__init__(key)
        self.feature_sampler = feature_sampler
        self.init_cache()

    def compute_features(self, state, resample=False):
        state_id = state.get_idx()
        if resample or state_id not in self.cache_state_idx_to_image:
            self.cache_state_idx_to_image[state_id] = self.feature_sampler(state.get_class())
        return self.cache_state_idx_to_image[state_id]

    def init_cache(self):
        self.cache_state_idx_to_image = {}


class ImageDiscretizer:
    def __init__(self, img, h_cells, w_cells, aug_cells=(0,0)):
        assert len(img.shape) == 3
        self.h_cells = h_cells
        self.w_cells = w_cells
        self.aug_cell_cnt_h, self.aug_cell_cnt_w = aug_cells
        self.cell_height = int(img.shape[0] // self.h_cells)
        self.cell_width = int(img.shape[1] // self.w_cells)
        self.img_clipped = img[:self.cell_height * self.h_cells, :self.cell_width * self.w_cells]
        self.img_clipped = np.pad(self.img_clipped,
                                  ((self.aug_cell_cnt_h*self.cell_height, self.aug_cell_cnt_h*self.cell_height),
                                   (self.aug_cell_cnt_w*self.cell_width, self.aug_cell_cnt_w*self.cell_width), (0,0)),
                                  mode="constant", constant_values=0)
        print("Note: image clipped to: {}".format(self.img_clipped.shape))
        self.img_lst = self._create_img_grid()
        self.idxs = np.arange(h_cells * w_cells).tolist()

    def _create_img_grid(self):
        img_lst = []
        for i in range(self.h_cells):
            img_lst.append([])
            for j in range(self.w_cells):
                img_lst[-1].append(self.img_clipped[ (i) * self.cell_height: (i +  2*self.aug_cell_cnt_h + 1) * self.cell_height,
                                   (j) * self.cell_width: (j + 2*self.aug_cell_cnt_w + 1) * self.cell_width])
        return np.asarray(img_lst)

    def transform_state_list(self, state_list):
        new_state_list = []
        for state in state_list:
            new_state_list.append((int(np.floor(state[0] / self.cell_width)),
                                   int(np.floor(state[1] / self.cell_height))))
        return new_state_list

    def transform_state_list_list(self, state_list_list):
        return [self.transform_state_list(state_list) for state_list in state_list_list]

    def get_raw_image(self):
        return self.img_clipped

    def get_image_grid(self):
        return self.img_lst

    def get_image_cell(self, row, col):
        return self.img_lst[row][col]

    def __call__(self):
        return self.get_image_grid()


class FeatureStateIdxToArray(AbstractStateFeatureSpec):

    def __init__(self, state_idx_to_array_fn, key=None):
        super().__init__(key)
        self.state_idx_to_array_fn = state_idx_to_array_fn

    def compute_features(self, state):
        return self.state_idx_to_array_fn(state.get_idx())

class FeatureStateLocToArray(AbstractStateFeatureSpec):

    def __init__(self, state_loc_to_array_fn, key=None):
        super().__init__(key)
        self.state_loc_to_array_fn = state_loc_to_array_fn

    def compute_features(self, state):
        return self.state_loc_to_array_fn(*state.get_location())


class FeatureClassIdxToArray(AbstractStateFeatureSpec):

    def __init__(self, class_idx_to_array_fn, key=None):
        super().__init__(key)
        self.class_idx_to_array_fn = class_idx_to_array_fn

    def compute_features(self, state):
        return self.class_idx_to_array_fn(state.get_class())



import itertools
import numpy as np
import navigation_mdp as NvMDP
"""
Guacamole
"""

class DiscreteLfDWorld2D:
    ''' Discrete World specification class '''

    def __init__(self, discrete_state_space, phi_spec, r_spec, slip_prob=0.,
                 s_lst_lst=None, a_lst_lst=None, class_ids=None):
        self.S = discrete_state_space
        self.phi_spec = phi_spec
        self.r_spec = r_spec
        self.class_ids = class_ids
        self.T = NvMDP.dynamics.XYDynamics(self.S, slip_prob=slip_prob)
        self.s_lst_lst = []
        self.a_lst_lst = []
        self.tau_lst = []
        self.attach_classes(class_ids)
        self.attach_feature_spec(phi_spec)
        self.attach_reward_spec(r_spec)
        self.add_trajectories(s_lst_lst, a_lst_lst)

    def attach_classes(self, class_ids):
        self.class_ids = class_ids
        self.S.attach_classes(class_ids)

    def clear_feature_spec(self):
        self.S.clear_feature_spec()

    def attach_feature_spec(self, phi_spec):
        self.phi_spec = phi_spec
        if isinstance(phi_spec, list):
            for __phi_spec in phi_spec:
                assert isinstance(__phi_spec, NvMDP.features.AbstractStateFeatureSpec), \
                    "Invalid feature spec. Supported type: {}".format(NvMDP.features.AbstractStateFeatureSpec)
                self.S.attach_feature_spec(__phi_spec)
        else:
            assert isinstance(phi_spec, NvMDP.features.AbstractStateFeatureSpec), \
                "Invalid feature spec. Supported type: {}".format(NvMDP.features.AbstractStateFeatureSpec)
            self.S.attach_feature_spec(phi_spec)

    def clear_reward_spec(self):
        self.S.clear_reward_spec()

    def attach_reward_spec(self, r_spec):
        self.r_spec = r_spec
        if isinstance(r_spec, list):
            for __r_spec in r_spec:
                assert isinstance(__r_spec, NvMDP.reward.AbstractStateRewardSpec), \
                    "Invalid reward spec. Supported type: {}".format(NvMDP.reward.AbstractStateRewardSpec)
                self.S.attach_reward_spec(__r_spec)
        else:
            assert isinstance(r_spec, NvMDP.reward.AbstractStateRewardSpec), \
                "Invalid reward spec. Supported type: {}".format(NvMDP.reward.AbstractStateRewardSpec)
            self.S.attach_reward_spec(r_spec)

    def state_space(self):
        return self.S

    def features(self, loc=None, idx=None, gridded=False, numpyize=True, key=None):
        return self.S.features(loc=loc, idx=idx, gridded=gridded, numpyize=numpyize, key=key)

    def dim_features(self, key=None):
        return self.S[0].get_features(key=key).shape

    def actions(self):
        return self.dynamics().actions()

    def rewards(self, numpyize=True, gridded=False, key=None):
        return self.S.rewards(numpyize=numpyize, gridded=gridded, key=key)

    def dynamics(self):
        return self.T

    def trajectories(self, s_a_zipped=False):
        if s_a_zipped:
            return self.tau_lst
        else:
            return self.s_lst_lst, self.a_lst_lst

    def clear_trajectories(self):
        self.s_lst_lst = []
        self.a_lst_lst = []
        self.tau_lst = []

    def _update_tau_lst(self):
        self.tau_lst = self._get_tau_lst(self.s_lst_lst, self.a_lst_lst)

    def _get_tau_lst(self, s_lst_lst, a_lst_lst):
        return [list(zip(s_lst_lst[i], a_lst_lst[i])) for i in range(len(s_lst_lst))]

    def add_trajectory(self, s_lst, a_lst=None):
        if s_lst is not None:
            if a_lst is None:
                a_lst = self.T.loc_lst_to_a_lst(s_lst)
            self.s_lst_lst.append(s_lst)
            self.a_lst_lst.append(a_lst)
            self._update_tau_lst()

    def add_trajectories(self, s_lst_lst, a_lst_lst=None):
        if s_lst_lst is not None:
            if a_lst_lst is None:
                a_lst_lst = []
                for s_lst in s_lst_lst:
                    a_lst_lst.append(self.T.loc_lst_to_a_lst(s_lst))
            for s_lst, a_lst in zip(s_lst_lst, a_lst_lst):
                self.s_lst_lst.append(s_lst)
                self.a_lst_lst.append(a_lst)
            self._update_tau_lst()

    def to_grid(self, values):
        return self.S._organize_to_grid(values)

    def state_dims(self):
        return self.S.shape()

    def get_feture_shape(self, key):
        return self.S[0].get_features(key=key).shape

    def get_reward_model(self, key):
        return self.S[0].get_reward_spec(key).r_model

    def reward(self, state, key):
        return self.S[0].get_reward_spec(key)(state)

    def preprocess(self, features, reward_key):
        return self.S[0].get_reward_spec(reward_key).preprocess(features)

    def preprocessed_features_lst(self, f_key, r_key):
        return self.preprocess(
            self.features(numpyize=False, key=f_key), r_key)

    def postprocess(self, rewards, reward_key):
        return self.S[0].get_reward_spec(reward_key).postprocess(rewards)



import numpy as np
# Navigation Views
import navigation_vis.Raster as NavGridView
from mpl_toolkits.axes_grid1 import make_axes_locatable
from matplotlib import cm as cm, pyplot as plt, colors as mplotcolors


def plot_grid_data_helper(data, ann=None, ann_sz=12, ann_col="black", title=None, grid=None, cmap=cm.viridis, **kwargs):
    p = NavGridView.Raster(data, ax=plt.gca()).render(cmap=cmap, **kwargs).ticks(minor=False)
    if ann is not None:
        p.show_cell_text(ann, fontsize=ann_sz, color_cb=lambda x: ann_col)
    if grid is not None:
        p.grid()
    if title is not None:
        p.title(title)
    return p

def highlight_cell(x, y, ax=None, **kwargs):
    rect = plt.Rectangle((x-.5, y-.5), 1, 1, fill=False, **kwargs)
    ax = ax or plt.gca()
    ax.add_patch(rect)
    return rect

def plot_irl_world(S, s_lst_lst=[], titles=["States", "Classes", "Features", "Rewards"],
                   figsize=(18, 12), cbar_pad=1.0, cbar_size="10%",
                   r_key=None, phi_key=None, r_round_to=3, clean_title=False,
                   v_range=[None, None, None, None]):
    state_title, class_title, feature_title, reward_title = titles
    v_range = [(None, None) if v is None else v for v in v_range]
    state_range, class_range, feature_range, reward_range = v_range
    if not clean_title:
        feature_title += "(type={})".format(phi_key)
        reward_title += "(type={})".format(r_key)

    plt.figure(figsize=figsize)
    p = NavGridViewPlotter(S, r_key=r_key, phi_key=phi_key)
    plt.subplot(2, 2, 1)
    p.plot_states(
        cmap=cm.viridis, ann_col="white",
        title=state_title, vmin=state_range[0], vmax=state_range[1]).colorbar(
        where="left", pad=cbar_pad, size=cbar_size).grid().add_pixel_trajectories(
        s_lst_lst, arrow_props={"lw": 3, "color": "black", "shrinkB": 10, "shrinkA": 10})
    plt.subplot(2, 2, 2)
    p.plot_classes(
        cmap=cm.viridis, ann_col="white",
        title=class_title, vmin=class_range[0], vmax=class_range[1]).colorbar(
        where="right", pad=cbar_pad, size=cbar_size).grid()
    plt.subplot(2, 2, 3)
    p.plot_features(
        ann=S.idxs.flatten(), cmap=None, ann_col="white",
        title=feature_title, vmin=feature_range[0], vmax=feature_range[1]).colorbar(
        where="left", pad=cbar_pad, size=cbar_size).grid().add_trajectories(
        s_lst_lst, arrow_props={"lw": 3, "color": "black", "shrinkB": 10, "shrinkA": 10})
    plt.subplot(2, 2, 4)
    p.plot_array(
        S.rewards(numpyize=True, key=r_key).round(r_round_to),
        cmap=cm.Blues_r, title=reward_title, vmin=reward_range[0], vmax=reward_range[1]).colorbar(
        where="right", pad=cbar_pad, size=cbar_size)
    plt.tight_layout()


def plot_irl_results(S, s_lst_lst, values, loglik_hist,
                     titles=["States", "Features", "Rewards", "Values", "Training Performance"],
                     figsize=(24, 24), cbar_pad=1.0, cbar_size="10%",
                     r_key=None, phi_key=None, r_round_to=3, clean_title=False,
                     v_range=[None, None, None, None, None], learned_lst_lst=[]):
    state_title, feature_title, reward_title, value_title, perf_title = titles
    v_range = [(None, None) if v is None else v for v in v_range]
    state_range, feature_range, reward_range, value_range, _ = v_range
    if not clean_title:
        feature_title += "(type={})".format(phi_key)
        reward_title += "(type={})".format(r_key)

    plt.figure(figsize=figsize)
    p = NavGridViewPlotter(S, r_key=r_key, phi_key=phi_key)
    plt.subplot(3, 2, 1)
    p.plot_states(
        cmap=cm.viridis, ann_col="white",
        title=state_title, vmin=state_range[0], vmax=state_range[1]).colorbar(
        where="left", pad=cbar_pad, size=cbar_size).grid().add_pixel_trajectories(
        s_lst_lst, arrow_props={"lw": 3, "color": "black", "shrinkB": 10, "shrinkA": 10})
    plt.subplot(3, 2, 2)
    p.plot_features(
        ann=S.idxs.flatten(), cmap=cm.viridis, ann_col="white",
        title=feature_title, vmin=feature_range[0], vmax=feature_range[1]).colorbar(
        where="right", pad=cbar_pad, size=cbar_size).grid().add_trajectories(
        s_lst_lst, arrow_props={"lw": 3, "color": "black", "shrinkB": 10, "shrinkA": 10})
    plt.subplot(3, 2, 3)
    p.plot_array(
        S.rewards(numpyize=True, key=r_key).round(r_round_to),
        cmap=cm.Blues_r, title=reward_title, vmin=reward_range[0], vmax=reward_range[1]).colorbar(
        where="left", pad=cbar_pad, size=cbar_size).add_pixel_trajectories(
        learned_lst_lst, arrow_props={"lw": 3, "color": "black", "shrinkB": 10, "shrinkA": 10})
    plt.subplot(3, 2, 4)
    p.plot_array(
        values, cmap=cm.Blues_r, title=value_title, vmin=value_range[0], vmax=value_range[1]).colorbar(
        where="right", pad=cbar_pad, size=cbar_size).add_pixel_trajectories(
        learned_lst_lst, arrow_props={"lw": 3, "color": "black", "shrinkB": 10, "shrinkA": 10})
    plt.subplot(3, 2, 5)
    plt.plot(list(range(len(loglik_hist))), loglik_hist)
    plt.xlabel("Iteration")
    plt.ylabel("Likelihood")
    plt.title(perf_title)
    plt.tight_layout()

class NavGridViewPlotter:
    def __init__(self, S, R=None, cartesian=False, r_key=None, phi_key=None):
        self.S = S
        self.R = S.rewards(key=r_key) if R is None else R
        self.PHI_gridded = self.S.features(gridded=True, key=phi_key)
        self.R_grided = self.S._organize_to_grid(self.R)
        self.class_ids_grided = self.S._organize_to_grid(self.S.class_ids)
        self.idxs_gridded = self.S.idxs
        self.cartesian = cartesian
        self.p = None

    def highlight_terminal_states(self):
        for s in self.S.get_terminal_states():
            r, c = s.location
            highlight_cell(c, r, ax=self.p.ax, color="white", linewidth=5)

    def plot_rewards(self, title="Rewards", *args, **kwargs):
        data = self.R_grided[..., np.newaxis, np.newaxis, np.newaxis]
        ann = self.R
        self.p = plot_grid_data_helper(data, ann, title=title, *args, **kwargs)
        if self.cartesian:
            self.p.ax.invert_yaxis()
        return self.p

    def plot_states(self, title="States", *args, **kwargs):
        data = self.idxs_gridded[..., np.newaxis, np.newaxis, np.newaxis]
        ann = self.S.idxs.flatten()
        self.p = plot_grid_data_helper(data, ann, title=title, *args, **kwargs)
        self.highlight_terminal_states()
        if self.cartesian:
            self.p.ax.invert_yaxis()
        return self.p

    def plot_array(self, data, title="Data", *args, **kwargs):
        data = self.S._organize_to_grid(np.asarray(data).flatten())[..., np.newaxis, np.newaxis, np.newaxis]
        ann = data.flatten()
        self.p = plot_grid_data_helper(data, ann, title=title, *args, **kwargs)
        if self.cartesian:
            self.p.ax.invert_yaxis()
        return self.p

    def plot_classes(self, title="Classes", *args, **kwargs):
        data = self.class_ids_grided[..., np.newaxis, np.newaxis, np.newaxis]
        ann = self.S.class_ids
        self.p = plot_grid_data_helper(data, ann, title=title, *args, **kwargs)
        self.highlight_terminal_states()
        if self.cartesian:
            self.p.ax.invert_yaxis()
        return self.p

    def plot_features(self, ann=None, title="Features", *args, **kwargs):
        if self.cartesian:
            data = NavGridView.flip_y_axis(self.PHI_gridded)
        else:
            data = self.PHI_gridded
        n_dim = len(data.shape)
        if n_dim == 3: # one-hot features
            H, W, K = data.shape
            if K < 10:
                data = data[..., np.newaxis, np.newaxis]
            else:
                # k1 = int(np.ceil(np.sqrt(K)))
                # k2 = int(np.ceil(K / k1))
                try:
                    k1 = H
                    k2 = W
                    data = data.reshape(H, W, k1, k2)[..., np.newaxis]
                except:
                    data = data.reshape(H, W, K, 1)[..., np.newaxis]
        elif n_dim == 4:
            data = data[..., np.newaxis]
        elif n_dim == 5:
            pass
        else:
            raise Exception("data dimension {} not supported!".format(n_dim))

        if ann is None:
            ann = self.S.class_ids
        ann = np.asarray(ann).flatten()
        self.p = plot_grid_data_helper(data, ann, title=title, *args, **kwargs)
        if self.cartesian:
            self.p.ax.invert_yaxis()
        return self.p

    def add_colorbar(self, *args, **kwargs):
        self.p.colorbar(*args, **kwargs)
        return self

    def add_trajectories(self, *args, ** kwargs):
        self.p.add_trajectories(*args, **kwargs)
        return self
